@@ -1,4 +1,4 @@
-/* Copyright (c) 2006, 2015, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2006, 2017, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -77,7 +77,8 @@ Relay_log_info::Relay_log_info(bool is_slave_recovery
    cur_log_fd(-1), relay_log(&sync_relaylog_period),
    is_relay_log_recovery(is_slave_recovery),
    save_temporary_tables(0),
-   cur_log_old_open_count(0), group_relay_log_pos(0), event_relay_log_pos(0),
+   cur_log_old_open_count(0), error_on_rli_init_info(false),
+   group_relay_log_pos(0), event_relay_log_pos(0),
    group_master_log_pos(0),
    gtid_set(global_sid_map, global_sid_lock),
    log_space_total(0), ignore_log_space_limit(0),
@@ -100,7 +101,7 @@ Relay_log_info::Relay_log_info(bool is_slave_recovery
    mts_group_status(MTS_NOT_IN_GROUP), reported_unsafe_warning(false),
    rli_description_event(NULL),
    sql_delay(0), sql_delay_end(0), m_flags(0), row_stmt_start_timestamp(0),
-   long_find_row_note_printed(false), error_on_rli_init_info(false)
+   long_find_row_note_printed(false)
 {
   DBUG_ENTER("Relay_log_info::Relay_log_info");
 
@@ -141,6 +142,7 @@ Relay_log_info::Relay_log_info(bool is_slave_recovery
   relay_log.init_pthread_objects();
   do_server_version_split(::server_version, slave_version_split);
   last_retrieved_gtid.clear();
+  force_flush_postponed_due_to_split_trans= false;
   DBUG_VOID_RETURN;
 }
 
@@ -220,7 +222,8 @@ void Relay_log_info::reset_notified_relay_log_change()
    @param shift          number of bits to shift by Worker due to the
                          current checkpoint change.
    @param new_ts         new seconds_behind_master timestamp value
-                         unless zero. Zero could be due to FD event.
+                         unless zero. Zero could be due to FD event
+                         or fake rotate event.
    @param need_data_lock False if caller has locked @c data_lock
 */
 void Relay_log_info::reset_notified_checkpoint(ulong shift, time_t new_ts,
@@ -307,15 +310,33 @@ bool Relay_log_info::mts_finalize_recovery()
     even temporary holes. Therefore stale records are deleted
     from the tail.
   */
+  DBUG_EXECUTE_IF("enable_mts_wokrer_failure_in_recovery_finalize",
+                  {DBUG_SET("+d,mts_worker_thread_init_fails");});
+
   for (i= recovery_parallel_workers; i > workers.elements && !ret; i--)
   {
     Slave_worker *w=
       Rpl_info_factory::create_worker(repo_type, i - 1, this, true);
-    ret= w->remove_info();
-    delete w;
+    /*
+      If an error occurs during the above create_worker call, the newly created
+      worker object gets deleted within the above function call itself and only
+      NULL is returned. Hence the following check has been added to verify
+      that a valid worker object exists.
+    */
+    if (w)
+    {
+      ret= w->remove_info();
+      delete w;
+    }
+    else
+    {
+      ret= true;
+      goto err;
+    }
   }
   recovery_parallel_workers= slave_parallel_workers;
 
+err:
   DBUG_RETURN(ret);
 }
 
@@ -397,7 +418,7 @@ void Relay_log_info::clear_until_condition()
   acquired it.
   @param errmsg[out] On error, this function will store a pointer to
   an error message here
-  @param look_for_description_event[in] If true, this function will
+  @param keep_looking_for_fd[in] If true, this function will
   look for a Format_description_log_event.  We only need this when the
   SQL thread starts and opens an existing relay log and has to execute
   it (possibly from an offset >4); then we need to read the first
@@ -412,7 +433,7 @@ void Relay_log_info::clear_until_condition()
 int Relay_log_info::init_relay_log_pos(const char* log,
                                        ulonglong pos, bool need_data_lock,
                                        const char** errmsg,
-                                       bool look_for_description_event)
+                                       bool keep_looking_for_fd)
 {
   DBUG_ENTER("Relay_log_info::init_relay_log_pos");
   DBUG_PRINT("info", ("pos: %lu", (ulong) pos));
@@ -501,7 +522,7 @@ int Relay_log_info::init_relay_log_pos(const char* log,
   if (pos > BIN_LOG_HEADER_SIZE) /* If pos<=4, we stay at 4 */
   {
     Log_event* ev;
-    while (look_for_description_event)
+    while (keep_looking_for_fd)
     {
       /*
         Read the possible Format_description_log_event; if position
@@ -542,23 +563,27 @@ int Relay_log_info::init_relay_log_pos(const char* log,
           describes the whole relay log; indeed, one can have this sequence
           (starting from position 4):
           Format_desc (of slave)
+          Previous-GTIDs (of slave IO thread, if GTIDs are enabled)
           Rotate (of master)
           Format_desc (of master)
           So the Format_desc which really describes the rest of the relay log
-          is the 3rd event (it can't be further than that, because we rotate
+          can be the 3rd or the 4th event (depending on GTIDs being enabled or
+          not, it can't be further than that, because we rotate
           the relay log when we queue a Rotate event from the master).
           But what describes the Rotate is the first Format_desc.
           So what we do is:
           go on searching for Format_description events, until you exceed the
-          position (argument 'pos') or until you find another event than Rotate
-          or Format_desc.
+          position (argument 'pos') or until you find an event other than
+          Previous-GTIDs, Rotate or Format_desc.
         */
       }
       else
       {
         DBUG_PRINT("info",("found event of another type=%d",
                            ev->get_type_code()));
-        look_for_description_event= (ev->get_type_code() == ROTATE_EVENT);
+        keep_looking_for_fd=
+          (ev->get_type_code() == ROTATE_EVENT ||
+           ev->get_type_code() == PREVIOUS_GTIDS_LOG_EVENT);
         delete ev;
       }
     }
@@ -604,7 +629,7 @@ err:
   @param[in]  log_name        log name to wait for,
   @param[in]  log_pos         position to wait for,
   @param[in]  timeout         @c timeout in seconds before giving up waiting.
-                              @c timeout is longlong whereas it should be ulong; but this is
+                              @c timeout is double whereas it should be ulong; but this is
                               to catch if the user submitted a negative timeout.
 
   @retval  -2   improper arguments (log_pos<0)
@@ -618,7 +643,7 @@ err:
 
 int Relay_log_info::wait_for_pos(THD* thd, String* log_name,
                                     longlong log_pos,
-                                    longlong timeout)
+                                    double timeout)
 {
   int event_count = 0;
   ulong init_abort_pos_wait;
@@ -633,7 +658,7 @@ int Relay_log_info::wait_for_pos(THD* thd, String* log_name,
   DBUG_PRINT("enter",("log_name: '%s'  log_pos: %lu  timeout: %lu",
                       log_name->c_ptr_safe(), (ulong) log_pos, (ulong) timeout));
 
-  set_timespec(abstime,timeout);
+  set_timespec_nsec(abstime, (ulonglong) timeout * 1000000000ULL);
   mysql_mutex_lock(&data_lock);
   thd->ENTER_COND(&data_cond, &data_lock,
                   &stage_waiting_for_the_slave_thread_to_advance_position,
@@ -813,7 +838,7 @@ improper_arguments: %d  timed_out: %d",
   /Alfranio
 */
 int Relay_log_info::wait_for_gtid_set(THD* thd, String* gtid,
-                                      longlong timeout)
+                                      double timeout)
 {
   int event_count = 0;
   ulong init_abort_pos_wait;
@@ -825,10 +850,10 @@ int Relay_log_info::wait_for_gtid_set(THD* thd, String* gtid,
   if (!inited)
     DBUG_RETURN(-2);
 
-  DBUG_PRINT("info", ("Waiting for %s timeout %lld", gtid->c_ptr_safe(),
+  DBUG_PRINT("info", ("Waiting for %s timeout %f", gtid->c_ptr_safe(),
              timeout));
 
-  set_timespec(abstime, timeout);
+  set_timespec_nsec(abstime, (ulonglong) timeout * 1000000000ULL);
   mysql_mutex_lock(&data_lock);
   thd->ENTER_COND(&data_cond, &data_lock,
                   &stage_waiting_for_the_slave_thread_to_advance_position,
@@ -1060,6 +1085,8 @@ int Relay_log_info::purge_relay_logs(THD *thd, bool just_reset,
                                      const char** errmsg)
 {
   int error=0;
+  const char *ln;
+  char name_buf[FN_REFLEN];
   DBUG_ENTER("Relay_log_info::purge_relay_logs");
 
   /*
@@ -1084,11 +1111,41 @@ int Relay_log_info::purge_relay_logs(THD *thd, bool just_reset,
   if (!inited)
   {
     DBUG_PRINT("info", ("inited == 0"));
-    DBUG_RETURN(0);
-  }
+    if (error_on_rli_init_info)
+    {
+      ln= relay_log.generate_name(opt_relay_logname, "-relay-bin",
+				  1, name_buf);
+      if (relay_log.open_index_file(opt_relaylog_index_name, ln, TRUE))
+      {
+	sql_print_error("Unable to purge relay log files. Failed to open relay "
+			"log index file:%s.", relay_log.get_index_fname());
 
-  DBUG_ASSERT(slave_running == 0);
-  DBUG_ASSERT(mi->slave_running == 0);
+	DBUG_RETURN(1);
+      }
+      mysql_mutex_lock(&mi->data_lock);
+      if (relay_log.open_binlog(ln, 0, SEQ_READ_APPEND,
+				(max_relay_log_size ? max_relay_log_size :
+				 max_binlog_size), true,
+				true/*need_lock_log=true*/,
+				true/*need_lock_index=true*/,
+				true/*need_sid_lock=true*/,
+				mi->get_mi_description_event()))
+      {
+	sql_print_error("Unable to purge relay log files. Failed to open relay "
+			"log file:%s.", relay_log.get_log_fname());
+        mysql_mutex_unlock(&mi->data_lock);
+	DBUG_RETURN(1);
+      }
+      mysql_mutex_unlock(&mi->data_lock);
+    }
+    else
+      DBUG_RETURN(0);
+  }
+  else
+  {
+    DBUG_ASSERT(slave_running == 0);
+    DBUG_ASSERT(mi->slave_running == 0);
+  }
 
   slave_skip_counter= 0;
   mysql_mutex_lock(&data_lock);
@@ -1129,6 +1186,11 @@ int Relay_log_info::purge_relay_logs(THD *thd, bool just_reset,
                               group_relay_log_pos,
                               false/*need_data_lock=false*/, errmsg, 0);
 
+  if (!inited && error_on_rli_init_info)
+    relay_log.close(LOG_CLOSE_INDEX | LOG_CLOSE_STOP_EVENT,
+                    true/*need_lock_log=true*/,
+                    true/*need_lock_index=true*/);
+
 err:
 #ifndef DBUG_OFF
   char buf[22];
@@ -1320,9 +1382,7 @@ bool Relay_log_info::is_until_satisfied(THD *thd, Log_event *ev)
     break;
 
   case UNTIL_SQL_AFTER_MTS_GAPS:
-#ifndef DBUG_OFF
   case UNTIL_DONE:
-#endif
     /*
       TODO: this condition is actually post-execution or post-scheduling
             so the proper place to check it before SQL thread goes
@@ -1337,9 +1397,7 @@ bool Relay_log_info::is_until_satisfied(THD *thd, Log_event *ev)
                             "UNTIL SQL_AFTER_MTS_GAPS as it has "
                             "processed all gap transactions left from "
                             "the previous slave session.");
-#ifndef DBUG_OFF
       until_condition= UNTIL_DONE;
-#endif
       DBUG_RETURN(true);
     }
     else
@@ -1798,6 +1856,7 @@ a file name for --relay-log-index option.", opt_relaylog_index_name);
     if (relay_log.open_binlog(ln, 0, SEQ_READ_APPEND,
                               (max_relay_log_size ? max_relay_log_size :
                                max_binlog_size), true,
+                              true/*need_lock_log=true*/,
                               true/*need_lock_index=true*/,
                               true/*need_sid_lock=true*/,
                               mi->get_mi_description_event()))
@@ -1897,7 +1956,11 @@ a file name for --relay-log-index option.", opt_relaylog_index_name);
     goto err;
   }
 
-  is_relay_log_recovery= FALSE;
+  /*
+    In case of MTS the recovery is deferred until the end of global_init_info.
+  */
+  if (!mi->rli->mts_recovery_group_cnt)
+    is_relay_log_recovery= FALSE;
   DBUG_RETURN(error);
 
 err:
@@ -1906,7 +1969,9 @@ err:
   error_on_rli_init_info= true;
   if (msg)
     sql_print_error("%s.", msg);
-  relay_log.close(LOG_CLOSE_INDEX | LOG_CLOSE_STOP_EVENT);
+  relay_log.close(LOG_CLOSE_INDEX | LOG_CLOSE_STOP_EVENT,
+                  true/*need_lock_log=true*/,
+                  true/*need_lock_index=true*/);
   DBUG_RETURN(error);
 }
 
@@ -1927,7 +1992,9 @@ void Relay_log_info::end_info()
     cur_log_fd= -1;
   }
   inited = 0;
-  relay_log.close(LOG_CLOSE_INDEX | LOG_CLOSE_STOP_EVENT);
+  relay_log.close(LOG_CLOSE_INDEX | LOG_CLOSE_STOP_EVENT,
+                  true/*need_lock_log=true*/,
+                  true/*need_lock_index=true*/);
   relay_log.harvest_bytes_written(&log_space_total);
   /*
     Delete the slave's temporary tables from memory.
@@ -2029,9 +2096,10 @@ int Relay_log_info::flush_info(const bool force)
   if (write_info(handler))
     goto err;
 
-  if (handler->flush_info(force))
+  if (handler->flush_info(force || force_flush_postponed_due_to_split_trans))
     goto err;
 
+  force_flush_postponed_due_to_split_trans= false;
   DBUG_RETURN(0);
 
 err:
